from unittest.mock import Mock, MagicMock, patch
import pytest
from google.cloud import bigquery
import pandas as pd
from cru_dse_utils import (
    get_schema_from_bigquery,
    upload_dataframe_to_bigquery,
    download_from_bigquery_as_dataframe,
    query_bigquery_as_dataframe,
)


# Test upload_dataframe_to_bigquery for a successful request where everything works as expected.
@patch("cru_dse_utils.bigquery.get_google_authorized_session")
def test_get_schema_from_bigquery_success(mock_get_session):
    response = MagicMock()
    mock_response = {"schema": {"fields": [{"name": "field1", "type": "STRING"}]}}
    response.status_code = 200
    response.json.return_value = mock_response

    session = MagicMock()
    session.get.return_value = response
    mock_get_session.return_value = session

    project_id, dataset_id, table_id, secrete_name = (
        "test_project",
        "test_dataset",
        "test_table",
        "test_secrete",
    )

    result = get_schema_from_bigquery(project_id, dataset_id, table_id, secrete_name)

    print(f"Mocked session: {mock_get_session()}")
    print(f"Result: {result}")

    # Ensure the session get method was called with the right argument
    url = f"https://bigquery.googleapis.com/bigquery/v2/projects/{project_id}/datasets/{dataset_id}/tables/{table_id}"
    session.get.assert_called_once_with(url)

    assert result == mock_response["schema"]["fields"]


# Test upload_dataframe_to_bigquery for server error.
@patch("cru_dse_utils.bigquery.get_google_authorized_session")
def test_get_schema_from_bigquery_server_error(mock_get_session):
    error_code = 500
    mock_get_session.return_value = MagicMock(
        get=MagicMock(return_value=MagicMock(status_code=error_code))
    )
    project_id, dataset_id, table_id, secrete_name = (
        "test_project",
        "test_dataset",
        "test_table",
        "test_secrete",
    )

    result = get_schema_from_bigquery(project_id, dataset_id, table_id, secrete_name)

    assert result is None


# Test upload_dataframe_to_bigquery for auth error.
@patch("cru_dse_utils.bigquery.get_google_authorized_session")
def test_get_schema_from_bigquery_auth_error(mock_get_session):
    mock_get_session.return_value = None
    project_id, dataset_id, table_id, secrete_name = (
        "test_project",
        "test_dataset",
        "test_table",
        "test_secrete",
    )

    result = get_schema_from_bigquery(project_id, dataset_id, table_id, secrete_name)

    assert result is None


# Test upload_dataframe_to_bigquery for a successful request where everything works as expected.
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.get_google_credentials")
def test_upload_dataframe_to_bigquery(
    mock_get_google_credentials, mock_bigquery_client
):
    # create some mock objects for the tests
    mock_credentials = Mock()
    mock_df = Mock(spec=pd.DataFrame)
    mock_job = Mock()
    mock_client = Mock()
    mock_load_job_config = Mock()

    # setup the return values for the mock objects
    mock_get_google_credentials.return_value = mock_credentials
    mock_bigquery_client.return_value = mock_client
    mock_client.load_table_from_dataframe.return_value = mock_job
    mock_load_job_config.return_value = mock_load_job_config

    # run the function with the mock objects
    upload_dataframe_to_bigquery(
        "project_id",
        "dataset_id",
        "table_id",
        "secret_name",
        mock_df,
        {},
    )

    # check that the mock objects were used as expected
    mock_get_google_credentials.assert_called_once_with("secret_name")
    mock_bigquery_client.assert_called_once_with(credentials=mock_credentials)
    mock_client.load_table_from_dataframe.assert_called_once()
    mock_job.result.assert_called_once()


# Test upload_dataframe_to_bigquery the override of the job config was used.
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.get_google_credentials")
def test_upload_dataframe_to_bigquery_with_config_override(
    mock_get_google_credentials, mock_bigquery_client
):
    # create some mock objects for the tests
    mock_credentials = Mock()
    mock_df = Mock(spec=pd.DataFrame)
    mock_job = Mock()
    mock_client = Mock()
    mock_load_job_config = Mock(spec=bigquery.LoadJobConfig)

    # setup the return values for the mock objects
    mock_get_google_credentials.return_value = mock_credentials
    mock_bigquery_client.return_value = mock_client
    mock_client.load_table_from_dataframe.return_value = mock_job

    # run the function with the mock objects and a mock LoadJobConfig
    upload_dataframe_to_bigquery(
        "project_id",
        "dataset_id",
        "table_id",
        "secret_name",
        mock_df,
        {},
        job_config_override=mock_load_job_config,
    )

    # check that the mock LoadJobConfig was used
    mock_client.load_table_from_dataframe.assert_called_once_with(
        mock_df, "project_id.dataset_id.table_id", job_config=mock_load_job_config
    )


# Test upload_dataframe_to_bigquery for an unsuccessful attempt to get the credentials.
@patch("cru_dse_utils.bigquery.get_google_credentials")
@patch("cru_dse_utils.bigquery.logging.getLogger")
def test_upload_dataframe_to_bigquery_no_credentials(
    mock_getLogger, mock_get_google_credentials
):
    # setup the return values for the mock objects
    mock_get_google_credentials.return_value = None
    mock_logger = Mock()
    mock_df = Mock(spec=pd.DataFrame)
    mock_getLogger.return_value = mock_logger

    # run the function with the mock objects
    try:
        upload_dataframe_to_bigquery(
            "project_id",
            "dataset_id",
            "table_id",
            "secret_name",
            mock_df,
            {},
        )
    except ValueError as e:
        # check that the appropriate error was thrown
        assert str(e) == "Invalid Google Cloud credentials"
    else:
        pytest.fail("Expected ValueError was not raised")

    # check that the appropriate error was logged
    mock_getLogger.assert_called_once_with("primary_logger")
    mock_logger.error.assert_called_once_with(
        "Upload to BigQuery error with authorization error"
    )


# Test download_from_bigquery_as_dataframe for a successful request where everything works as expected.
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.get_google_credentials")
def test_download_from_bigquery_as_dataframe(
    mock_get_google_credentials, mock_bigquery_client
):
    # create some mock objects for the tests
    mock_credentials = Mock()
    mock_df = Mock(spec=pd.DataFrame)
    mock_list_rows_result = MagicMock()
    mock_client = Mock()

    # setup the return values for the mock objects
    mock_get_google_credentials.return_value = mock_credentials
    mock_bigquery_client.return_value = mock_client
    mock_client.list_rows.return_value = mock_list_rows_result
    mock_list_rows_result.to_dataframe.return_value = mock_df

    # run the function with the mock objects
    result = download_from_bigquery_as_dataframe(
        "project_id",
        "dataset_id",
        "table_id",
        "secret_name",
    )

    # check that the mock objects were used as expected
    mock_get_google_credentials.assert_called_once_with("secret_name")
    mock_bigquery_client.assert_called_once_with(credentials=mock_credentials)
    mock_client.list_rows.assert_called_once()
    mock_list_rows_result.to_dataframe.assert_called_once()
    assert isinstance(result, pd.DataFrame)


# Test download_from_bigquery_as_dataframe for an unsuccessful attempt to get the credentials.
@patch("cru_dse_utils.bigquery.get_google_credentials")
@patch("cru_dse_utils.bigquery.logging.getLogger")
def test_download_from_bigquery_as_dataframe_no_credentials(
    mock_getLogger, mock_get_google_credentials
):
    # setup the return values for the mock objects
    mock_get_google_credentials.return_value = None
    mock_logger = Mock()
    mock_getLogger.return_value = mock_logger

    # run the function with the mock objects

    result = download_from_bigquery_as_dataframe(
        "project_id",
        "dataset_id",
        "table_id",
        "secret_name",
    )

    assert result is None


# Test query_bigquery_as_dataframe for a successful request where everything works as expected.
@patch("cru_dse_utils.bigquery.get_google_credentials")
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.logging")
def test_query_bigquery_as_dataframe(mock_logging, mock_client, mock_get_credentials):
    # Arrange
    query = "SELECT * FROM `bigquery-public-data.samples.gsod`"
    secrete_name = "MY_SECRET"
    mock_get_credentials.return_value = "fake_credentials"
    mock_client_instance = mock_client.return_value
    mock_query_job = MagicMock()
    mock_client_instance.query.return_value = mock_query_job
    mock_results = MagicMock()
    mock_query_job.result.return_value = mock_results
    mock_results.to_dataframe.return_value = pd.DataFrame()

    # Act
    result = query_bigquery_as_dataframe(query, secrete_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secrete_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.query.assert_called_once_with(query)
    mock_query_job.result.assert_called_once()
    mock_results.to_dataframe.assert_called_once()
    assert isinstance(result, pd.DataFrame)


# Test query_bigquery_as_dataframe for an unsuccessful request where the credentials are invalid.
@patch("cru_dse_utils.bigquery.get_google_credentials")
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.logging")
def test_query_bigquery_as_dataframe_no_credentials(
    mock_logging, mock_client, mock_get_credentials
):
    # Arrange
    query = "SELECT * FROM `bigquery-public-data.samples.gsod`"
    secrete_name = "MY_SECRET"
    mock_get_credentials.return_value = None

    # Act
    result = query_bigquery_as_dataframe(query, secrete_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secrete_name)
    mock_logging.getLogger.return_value.error.assert_called_once_with(
        "Query BigQuery error with authorization error"
    )
    assert result is None


# Test query_bigquery_as_dataframe for an unsuccessful request where the query fails.
@patch("cru_dse_utils.bigquery.get_google_credentials")
@patch("cru_dse_utils.bigquery.bigquery.Client")
@patch("cru_dse_utils.bigquery.logging")
def test_query_bigquery_as_dataframe_query_error(
    mock_logging, mock_client, mock_get_credentials
):
    # Arrange
    query = "SELECT * FROM `bigquery-public-data.samples.gsod`"
    secrete_name = "MY_SECRET"
    mock_get_credentials.return_value = "fake_credentials"
    mock_client_instance = mock_client.return_value
    mock_client_instance.query.side_effect = Exception("Query failed")

    # Act
    result = query_bigquery_as_dataframe(query, secrete_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secrete_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.query.assert_called_once_with(query)
    mock_logging.getLogger.return_value.exception.assert_called_once()
    assert result is None



from .general import get_request
from .auth import (
    get_google_credentials,
    get_google_authorized_session,
    get_general_credentials,
)
from .bigquery import (
    get_schema_from_bigquery,
    upload_dataframe_to_bigquery,
    query_bigquery_as_dataframe,
    download_from_bigquery_as_dataframe,
)
from .gcs import upload_to_gcs, download_from_gcs_as_dataframe, upload_dataframe_to_gcs
from .dbt import get_dbt_job_list, trigger_dbt_job, get_dbt_run_status, dbt_run



from typing import List, Dict, Any, Optional, Union
import logging
import io
from google.cloud import storage
from cru_dse_utils import get_google_credentials
import pandas as pd


def upload_to_gcs(
    file_path: str, bucket_name: str, blob_name: str, secret_name: str
) -> None:
    """
    Uploads the specified file to a Google Cloud Storage bucket.

    This function uploads a file to the specified Google Cloud Storage bucket
    with the specified name. The function logs a message indicating whether
    the upload succeeded or failed.

    Args:
        file_path: str: The path to the file to be uploaded.
        bucket_name (str): The name of the Google Cloud Storage bucket to
        upload the file to.
        blob_name (str): The name of the file to create in the bucket.
        secret_name (str): The name of the environment variable to retrieve
        the Google Cloud credentials.

    Returns:
        None
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        logger.error(f"Failed to get Google Cloud credentials with {secret_name}")
        return None
    client = storage.Client(credentials=credentials)
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    try:
        blob.upload_from_filename(file_path)
        logger.info(
            f"Uploaded file to Google Cloud Storage: gs://{bucket_name}/{blob_name}"
        )
    except Exception as e:
        logger.exception(f"Upload to Google Cloud Storage error: {str(e)}")


def download_from_gcs_as_dataframe(
    bucket_name, blob_name, secret_name: str
) -> Union[pd.DataFrame, None]:
    """
    Downloads a file from a Google Cloud Storage bucket.

    This function downloads a file from the specified Google Cloud Storage
    bucket with the specified name. The function logs a message indicating
    whether the download succeeded or failed.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to
        download the file from.
        blob_name (str): The name of the file to download from the bucket.
        secret_name (str): The name of the environment variable to retrieve
        the Google Cloud credentials.

    Returns:
        pd.DataFrame: The contents of the file as a pandas DataFrame.
        None: If the download failed.
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        logger.error(f"Failed to get Google Cloud credentials with {secret_name}")
        return None
    client = storage.Client(credentials=credentials)
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    try:
        contents = blob.download_as_string()
        df = pd.read_csv(io.BytesIO(contents))
        logger.info(f"Downloaded data from gs://{bucket_name}/{blob_name}")
        return df
    except Exception as e:
        logger.exception(f"Download from Google Cloud Storage error: {str(e)}")


def upload_dataframe_to_gcs(
    bucket_name: str, blob_name: str, df: pd.DataFrame, secret_name: str
) -> None:
    """
    Uploads a pandas DataFrame to a Google Cloud Storage bucket.

    This function uploads a pandas DataFrame to the specified Google Cloud
    Storage bucket. The function logs a message indicating whether the upload
    succeeded or failed.

    Args:
        bucket_name (str): The name of the Google Cloud Storage bucket to
        upload the file to.
        blob_name (str): The name of the file to create in the bucket.
        df (pd.DataFrame): The pandas DataFrame to upload.
        secret_name (str): The name of the environment variable to retrieve
        the Google Cloud credentials.

    Returns:
        None
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        logger.error(f"Failed to get Google Cloud credentials with {secret_name}")
        return None
    client = storage.Client(credentials=credentials)
    bucket = client.bucket(bucket_name)
    blob = bucket.blob(blob_name)
    try:
        blob.upload_from_string(df.to_csv(index=False), "text/csv")
        logger.info(f"Uploaded data to gs://{bucket_name}/{blob_name}")
    except Exception as e:
        logger.exception(f"Upload to Google Cloud Storage error: {str(e)}")



from typing import List, Dict, Any, Optional, Union
import logging
import pandas as pd
from google.cloud import bigquery
from google.cloud.bigquery import SchemaField
from cru_dse_utils import get_google_credentials, get_google_authorized_session


def get_schema_from_bigquery(
    project_id: str, dataset_id: str, table_id: str, secret_name: str
) -> Optional[List[Dict]]:
    """
    Fetches and returns the schema details of the specified table in Google
    BigQuery.

    This function fetches the schema details by building the required URL
    using project_id, dataset_id, and table_id. The function checks for a
    successful session authorization and successful response status for a
    GET request. It also logs the progress and possible errors during the
    operation.

    Args:
        project_id (str): The Google BigQuery Project ID.
        dataset_id (str): The Google BigQuery Dataset ID.
        table_id (str): The Google BigQuery Table ID.

    Returns:
        List[Dict] or None: A list of dictionaries in json format containing
        the schema details if the operation is successful, None otherwise.
    """
    logger = logging.getLogger("primary_logger")
    session = get_google_authorized_session(secret_name)
    if session is None:
        logger.error("Get schema from BigQuery error with authorization error")
        return None
    url = f"https://bigquery.googleapis.com/bigquery/v2/projects/{project_id}/datasets/{dataset_id}/tables/{table_id}"
    response = session.get(url)
    if response.status_code == 200:
        data = response.json()
        schema_json = data["schema"]["fields"]
        logger.info("Retrieved schemas from BigQuery.")
        return schema_json
    else:
        logger.error(
            f"Get schema from BigQuery error with status code {response.status_code}, response: {response.text}"
        )
        return None


def validate_upload_dataframe_to_bigquery_arguments(
    project_id: str,
    dataset_id: str,
    table_id: str,
    secret_name: str,
    df: pd.DataFrame,
):
    """
    Validate the arguments needed for upload_bigquery_dataframe().

    Args:
        project_id (str): The Google Cloud project ID.
        dataset_id (str): The BigQuery dataset ID.
        table_id (str): The BigQuery table ID.
        secret_name (str): The name of the environment variable
        used for Google Cloud authentication.
        df (pd.DataFrame): The pandas DataFrame containing the data to
        upload.

    Raises:
        AssertionError: If any of the arguments are None or if df is not
        a pandas DataFrame.
    """
    assert project_id, "project_id must not be None or empty"
    assert dataset_id, "dataset_id must not be None or empty"
    assert table_id, "table_id must not be None or empty"
    assert secret_name, "secret_name must not be None or empty"
    assert isinstance(df, pd.DataFrame), "df must be a pandas DataFrame"


def get_job_config(schema_json, write_disposition, job_config_override):
    """
    Get the BigQuery LoadJobConfig based on provided arguments.

    Args:
        schema_json (Dict[Any, Any]): A dictionary representing the schema
        of the BigQuery table.
        write_disposition (str): Write disposition for the load job. Default
        is "WRITE_TRUNCATE", other options are "WRITE_APPEND" and
        "WRITE_EMPTY".
        job_config_override (bigquery.LoadJobConfig): An optional
        LoadJobConfig instance. If provided, this config will be used and
        other parameters will be ignored.

    Returns:
        bigquery.LoadJobConfig: The load job configuration for uploading
        data to BigQuery.
    """
    if job_config_override:
        return job_config_override
    if schema_json:
        schema = [SchemaField.from_api_repr(field) for field in schema_json]
        return bigquery.LoadJobConfig(
            schema=schema,
            create_disposition="CREATE_IF_NEEDED",
            write_disposition=write_disposition,
        )
    return bigquery.LoadJobConfig(
        autodetect=True,
        create_disposition="CREATE_IF_NEEDED",
        write_disposition=write_disposition,
    )


def upload_dataframe_to_bigquery(
    project_id: str,
    dataset_id: str,
    table_id: str,
    secret_name: str,
    df: pd.DataFrame,
    schema_json: Optional[Dict[Any, Any]] = None,
    job_config_override: Optional[bigquery.LoadJobConfig] = None,
    write_disposition: Optional[str] = "WRITE_TRUNCATE",
) -> None:
    """
    Uploads data from a pandas DataFrame to a BigQuery table.

    This function validates the provided arguments, retrieves the Google Cloud
    credentials, prepares the load job configuration and then starts the
    upload process.

    Args:
        project_id (str): The Google Cloud project ID.
        dataset_id (str): The BigQuery dataset ID.
        table_id (str): The BigQuery table ID.
        secret_name (str): The name of the environment variable used for
        Google Cloud authentication.
        df (pd.DataFrame): The pandas DataFrame containing the data to upload.
        write_disposition (str, optional): Write disposition for the load job.
        Default is "WRITE_TRUNCATE". Other options are "WRITE_APPEND" and
        "WRITE_EMPTY".
        schema_json (Dict[Any, Any], optional): A dictionary representing the
        schema of the BigQuery table.
        job_config_override (bigquery.LoadJobConfig, optional): An optional
        LoadJobConfig instance. If provided, this config will be used and
        other parameters will be ignored.

    Raises:
        ValueError: If Google Cloud credentials are invalid or absent.
        Any exception raised during the upload process will be re-raised after
        being logged.
    """
    validate_upload_dataframe_to_bigquery_arguments(
        project_id, dataset_id, table_id, secret_name, df
    )

    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        logger.error("Upload to BigQuery error with authorization error")
        raise ValueError("Invalid Google Cloud credentials")

    client = bigquery.Client(credentials=credentials)
    table_id_full = f"{project_id}.{dataset_id}.{table_id}"
    job_config = get_job_config(schema_json, write_disposition, job_config_override)

    logger.info(f"Starting to upload data to {table_id_full}")
    try:
        job = client.load_table_from_dataframe(df, table_id_full, job_config=job_config)
        job.result()
        logger.info(f"Uploaded data to {table_id_full}")
    except Exception as e:
        logger.exception(f"Upload to BigQuery error: {str(e)}")
        raise


def download_from_bigquery_as_dataframe(
    project_id: str, dataset_id: str, table_id: str, secret_name: str
) -> Union[pd.DataFrame, None]:
    """
    Download data from a BigQuery table to a pandas DataFrame.

    This function uploads a pandas DataFrame to the specified BigQuery table.
    The function logs a message indicating whether the upload succeeded or
    failed.

    Args:
        project_id (str): The Google Cloud project ID.
        dataset_id (str): The BigQuery dataset ID.
        table_id (str): The BigQuery table ID.
        secret_name (str): The name of the environment variable used for
        Google Cloud authentication.

    Returns:
        Union[pd.DataFrame, None]: The pandas DataFrame containing the data
        downloaded from BigQuery. If the authorization or download fails,
        None is returned.
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        logger.error("Download from BigQuery error with authorization error")
        return None
    client = bigquery.Client(credentials=credentials)
    table_id_full = f"{project_id}.{dataset_id}.{table_id}"
    logger.info(f"Starting to download data from BigQuery table: {table_id_full}")
    try:
        df = client.list_rows(table_id_full).to_dataframe()
        logger.info(f"Downloaded data from BigQuery table: {table_id_full}")
        return df
    except Exception as e:
        logger.exception(f"Download from BigQuery error: {str(e)}")
        return None


def query_bigquery_as_dataframe(
    query: str, secrete_name: str
) -> Union[pd.DataFrame, None]:
    """
    Executes a SQL query on a BigQuery dataset and returns the results as
    a pandas DataFrame.

    This function executes a SQL query on a BigQuery dataset and returns
    the results as a pandas DataFrame. The function logs a message indicating
    the query that was executed, and logs a message indicating the number of
    rows that were returned by the query.

    Args:
        query (str): The SQL query to execute.
        secrete_name (str): The name of the environment variable used for
        Google Cloud authentication.

    Returns:
        Union[pd.DataFrame, None]: A pandas DataFrame containing the results
        of the query. If the authorization or query fails, None is returned.
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secrete_name)
    if credentials is None:
        logger.error("Query BigQuery error with authorization error")
        return None
    client = bigquery.Client(credentials=credentials)
    try:
        query_job = client.query(query)
        results = query_job.result()
        logger.info(f"Executed query.")
        return results.to_dataframe()
    except Exception as e:
        logger.exception(f"Query BigQuery error: {str(e)}")
        return None



import logging
import requests
import time
from typing import List, Dict, Any, Optional, Union
from cru_dse_utils import get_general_credentials


def get_dbt_job_list(account_id: str, secret_name: str) -> None:
    """
    Fetches a list of dbt jobs for a given account.

    This function retrieves the dbt jobs list from the dbt Cloud API for a
    given account. The function logs a message indicating whether the
    operation was successful or failed. This function is intended to be
    used for testing dbt connections and permissions.

    Args:
        account_id (str): The ID of the account in dbt Cloud.
        secret_name (str): The name of the environment variable used for
        dbt Cloud API token.

    Returns:
        None.
    """
    logger = logging.getLogger("primary_logger")
    token = get_general_credentials(secret_name)
    if token is None:
        logger.error(f"Failed to get dbt token with {secret_name}")
        return None
    headers = {
        "Authorization": f"Token {token}",
        "Content-Type": "application/json",
    }
    url = f"https://cloud.getdbt.com/api/v2/accounts/{account_id}/jobs/"
    try:
        r = requests.get(url, headers=headers)
        r.raise_for_status()
        logger.info(f"Got dbt job list.")
    except Exception as e:
        logger.exception(f"List dbt job error: {str(e)}")


def trigger_dbt_job(account_id: str, job_id: str, token: str) -> Union[str, None]:
    """
    Triggers a dbt job.

    This function triggers a dbt job by sending a POST request to the
    specified URL with the specified headers and JSON payload. The function
    logs a message indicating whether the dbt job started successfully or
    failed, and returns the id of the dbt job run.

    Args:
        account_id (str): The ID of the account in dbt Cloud.
        job_id (str): The ID of the dbt job to be triggered.
        token (str): The dbt Cloud API token.

    Returns:
        str: The run_id of the dbt job run.
        None: If the dbt job failed to start.
    """
    logger = logging.getLogger("primary_logger")
    headers = {
        "Authorization": f"Token {token}",
        "Content-Type": "application/json",
    }
    url = f"https://cloud.getdbt.com/api/v2/accounts/{account_id}/jobs/{job_id}/run/"
    json = {"cause": "Triggered by python script API request"}
    try:
        r = requests.post(url, headers=headers, json=json)
        r.raise_for_status()
        logger.info(f"dbt job started successfully.")
        run_id = r.json()["data"]["id"]
        return run_id
    except Exception as e:
        logger.exception(f"dbt job failed to start: {str(e)}")
        return None


def get_dbt_run_status(run_id: str, token: str) -> int:
    """
    Retrieves the status of a dbt job run.

    This function retrieves the status of a dbt job run by sending a GET
    request to the specified URL with the specified headers and the dbt
    job run_id. The function logs a message indicating whether the dbt job
    status check succeeded or failed, and returns the status of the dbt
    job run.

    Args:
        run_id (str): The id of the dbt job run.
        token (str): The dbt Cloud API token.

    Returns:
        int: The status of the dbt job run.

    """
    logger = logging.getLogger("primary_logger")
    headers = {
        "Authorization": f"Token {token}",
        "Content-Type": "application/json",
    }
    url = f"https://cloud.getdbt.com/api/v2/accounts/10206/runs/{run_id}/"
    try:
        r = requests.get(url, headers=headers)
        r.raise_for_status()
        status = r.json()["data"]["status"]
    except Exception as e:
        logger.exception(f"dbt job status check failed: {e}")
        status = 0
    return status


def dbt_run(
    account_id: str, job_id: str, secret_name: str, max_retries: int = 3
) -> None:
    """
    Runs a dbt job and checks its status, with retry logic.

    This function runs a dbt job by calling the `trigger_dbt_job()` function
    and then checks the status of the job by calling `get_dbt_run_status()`
    function every 30 seconds until the job is completed successfully,
    failed, or cancelled. If the job fails, it will retry up to 'max_retries' times.
    The function logs messages indicating the status of the dbt job and its final outcome.

    Args:
        account_id (str): The ID of the account in dbt Cloud.
        job_id (str): The ID of the dbt job to be triggered.
        secret_name (str): The name of the environment variable used for dbt Cloud API token.
        max_retries (int): Maximum number of retries if the job fails. Default is 3.

    Returns:
        None
    """
    logger = logging.getLogger("primary_logger")
    logger.info(f"Starting dbt job...")
    token = get_general_credentials(secret_name)
    if token is None:
        logger.error(f"Failed to get dbt token with {secret_name}")
        return

    retries = 0
    while retries <= max_retries:
        run_id = trigger_dbt_job(account_id, job_id, token)
        if run_id is None:
            logger.error(
                f"dbt run failed to start. Retry {retries + 1} of {max_retries + 1}"
            )
            retries += 1
            continue

        while True:
            time.sleep(30)
            status = get_dbt_run_status(run_id, token)
            status_str = {
                0: "Status not available",
                1: "Queued",
                2: "Starting",
                3: "Running",
                10: "Success",
                20: "Failed",
                30: "Cancelled",
            }.get(status, "Unknown")

            logger.info(f"dbt job status: {status} - {status_str}")

            if status == 10:
                logger.info(f"dbt job run completed successfully.")
                return  # Exit the function on success
            elif status in (20, 30):  # Failed or Cancelled
                if retries < max_retries:
                    logger.warning(
                        f"dbt job {status_str}. Retrying... (Attempt {retries + 1} of {max_retries})"
                    )
                    retries += 1
                    break  # Break the inner loop to retry
                else:
                    logger.error(f"dbt job {status_str} after {max_retries} retries.")
                    return  # Exit the function after max retries

    logger.error(
        f"dbt job failed to complete successfully after {max_retries} retries."
    )



from unittest.mock import Mock, MagicMock, patch
import pytest
import pandas as pd
from cru_dse_utils import (
    upload_to_gcs,
    download_from_gcs_as_dataframe,
    upload_dataframe_to_gcs,
)


# Fixture to set up variables for testing.
@pytest.fixture
def setup_variables():
    file_path = "/path/to/file"
    bucket_name = "my-bucket"
    blob_name = "my-blob"
    secret_name = "MY_SECRET"
    return file_path, bucket_name, blob_name, secret_name


# Test upload_to_gcs for a successful request where everything works as expected.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
def test_upload_to_gcs(
    mock_logging, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = "fake_credentials"
    mock_client_instance = mock_client.return_value
    mock_bucket = MagicMock()
    mock_client_instance.bucket.return_value = mock_bucket
    mock_blob = MagicMock()
    mock_bucket.blob.return_value = mock_blob

    # Act
    upload_to_gcs(file_path, bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_blob.upload_from_filename.assert_called_once_with(file_path)


# Test upload_to_gcs for an unsuccessful attempt to retrieve the Google Cloud credentials.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
def test_upload_to_gcs_no_credentials(
    mock_logging, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = None

    # Act
    result = upload_to_gcs(file_path, bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_logging.getLogger.return_value.error.assert_called_once_with(
        f"Failed to get Google Cloud credentials with {secret_name}"
    )
    assert result is None


# Test upload_to_gcs for an unsuccessful attempt to upload the file to Google Cloud Storage.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
def test_upload_to_gcs_upload_error(
    mock_logging, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = "fake_credentials"
    mock_client_instance = mock_client.return_value
    mock_bucket = MagicMock()
    mock_client_instance.bucket.return_value = mock_bucket
    mock_blob = MagicMock()
    mock_blob.upload_from_filename.side_effect = Exception("Upload failed")
    mock_bucket.blob.return_value = mock_blob

    # Act
    result = upload_to_gcs(file_path, bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_blob.upload_from_filename.assert_called_once_with(file_path)
    mock_logging.getLogger.return_value.exception.assert_called_once()
    assert result is None


# Test download_from_gcs_as_dataframe for a successful request where everything works as expected.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
@patch("cru_dse_utils.gcs.pd.read_csv")
@patch("cru_dse_utils.gcs.io.BytesIO")
def test_download_from_gcs_as_dataframe(
    mock_bytes_io,
    mock_read_csv,
    mock_logging,
    mock_client,
    mock_get_credentials,
    setup_variables,
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = "fake_credentials"
    mock_bucket = MagicMock()
    mock_blob = MagicMock()
    mock_client_instance = mock_client.return_value
    mock_client_instance.bucket.return_value = mock_bucket
    mock_bucket.blob.return_value = mock_blob
    mock_content = MagicMock()
    mock_blob.download_as_string.return_value = mock_content
    mock_bytes_io.return_value = mock_content
    mock_dataframe = pd.DataFrame()
    mock_read_csv.return_value = mock_dataframe

    # Act
    returned_df = download_from_gcs_as_dataframe(bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_blob.download_as_string.assert_called_once()
    mock_bytes_io.assert_called_once_with(mock_content)
    mock_read_csv.assert_called_once_with(mock_content)
    assert isinstance(returned_df, pd.DataFrame)


# Test download_from_gcs_as_dataframe for an unsuccessful attempt to retrieve credentials.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
def test_download_from_gcs_as_dataframe_no_credentials(
    mock_logging, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = None

    # Act
    result = download_from_gcs_as_dataframe(bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_logging.getLogger.return_value.error.assert_called_once_with(
        f"Failed to get Google Cloud credentials with {secret_name}"
    )
    assert result is None


# Test download_from_gcs_as_dataframe for an unsuccessful attempt to download.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging")
def test_download_from_gcs_as_dataframe_download_error(
    mock_logging, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    mock_get_credentials.return_value = "fake_credentials"
    mock_client_instance = mock_client.return_value
    mock_bucket = MagicMock()
    mock_client_instance.bucket.return_value = mock_bucket
    mock_blob = MagicMock()
    mock_blob.download_as_string.side_effect = Exception("Download failed")
    mock_bucket.blob.return_value = mock_blob

    # Act
    result = download_from_gcs_as_dataframe(bucket_name, blob_name, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_logging.getLogger.return_value.exception.assert_called_once()
    assert result is None


# Test upload_dataframe_to_gcs for a successful request where everything works as expected.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging.getLogger")
def test_upload_dataframe_to_gcs_success(
    mock_get_logger, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    mock_get_credentials.return_value = "fake_credentials"
    mock_bucket = MagicMock()
    mock_blob = MagicMock()
    mock_client_instance = mock_client.return_value
    mock_client_instance.bucket.return_value = mock_bucket
    mock_bucket.blob.return_value = mock_blob
    mock_logger = MagicMock()
    mock_get_logger.return_value = mock_logger

    # Act
    upload_dataframe_to_gcs(bucket_name, blob_name, df, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_blob.upload_from_string.assert_called_once_with(
        df.to_csv(index=False), "text/csv"
    )
    mock_logger.info.assert_called_once_with(
        f"Uploaded data to gs://{bucket_name}/{blob_name}"
    )


# Test upload_dataframe_to_gcs for an unsuccessful attempt.
@patch("cru_dse_utils.gcs.get_google_credentials")
@patch("cru_dse_utils.gcs.storage.Client")
@patch("cru_dse_utils.gcs.logging.getLogger")
def test_upload_dataframe_to_gcs_fail(
    mock_get_logger, mock_client, mock_get_credentials, setup_variables
):
    # Arrange
    file_path, bucket_name, blob_name, secret_name = setup_variables
    df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    mock_get_credentials.return_value = "fake_credentials"
    mock_bucket = MagicMock()
    mock_blob = MagicMock()
    mock_blob.upload_from_string.side_effect = Exception("fake exception")
    mock_client_instance = mock_client.return_value
    mock_client_instance.bucket.return_value = mock_bucket
    mock_bucket.blob.return_value = mock_blob
    mock_logger = MagicMock()
    mock_get_logger.return_value = mock_logger

    # Act
    upload_dataframe_to_gcs(bucket_name, blob_name, df, secret_name)

    # Assert
    mock_get_credentials.assert_called_once_with(secret_name)
    mock_client.assert_called_once_with(credentials="fake_credentials")
    mock_client_instance.bucket.assert_called_once_with(bucket_name)
    mock_bucket.blob.assert_called_once_with(blob_name)
    mock_blob.upload_from_string.assert_called_once_with(
        df.to_csv(index=False), "text/csv"
    )
    mock_logger.exception.assert_called_once_with(
        "Upload to Google Cloud Storage error: fake exception"
    )



import os
import json
import base64
import logging
from typing import List, Dict, Any, Optional, Union
from google.oauth2.service_account import Credentials
from google.auth.transport.requests import AuthorizedSession


def get_google_credentials(secret_name: str) -> Union[Credentials, None]:
    """
    Retrieves Google Cloud credentials for the specified secret from local
    or Amazon ECS environment variables.

    This function retrieves the base64 encoded secrete value from environment
    variables by calling the `os.getenv()` function with the provided
    secrete name. The function then retrieves the Google Cloud credentials
    by calling the `from_service_account_info()` function with the service
    account info from The decoded secrete value. The function returns the
    resulting `Credentials` object.

    Args:
        secret_name (str): The name of the environment variable containing
        the secret value.

    Returns:
        A `Credentials` object or None: Retrived Google Cloud credentials.
        or None if failed to retrive credentials.
    """
    logger = logging.getLogger("primary_logger")
    try:
        secret_value_encoded = os.getenv(secret_name)
        if secret_value_encoded is None:
            logger.error(f"Failed to get Google Cloud credentials with {secret_name}")
            return None
        credentials_dict = json.loads(
            base64.b64decode(secret_value_encoded).decode("utf-8")
        )
        scopes = [
            "https://www.googleapis.com/auth/cloud-platform",
            "https://www.googleapis.com/auth/drive",
            "https://www.googleapis.com/auth/bigquery",
        ]
        credentials = Credentials.from_service_account_info(
            credentials_dict, scopes=scopes
        )
        return credentials
    except Exception as e:
        logger.exception(f"Get Google Cloud credentials error with {str(e)}")
        return None


def get_google_authorized_session(secret_name: str) -> Union[AuthorizedSession, None]:
    """
    Convert Google Cloud credentials to AuthorizedSession for the specified
    secret from local or Amazon ECS environment variables.

    This function calls the get_google_credentials() with the provided
    secrete name. The function then converts the resulting `Credentials`
    object to an `AuthorizedSession` object. The function returns the
    resulting `AuthorizedSession` object.

    Args:
        secret_name (str): The name of the environment variable containing
        the secret value.

    Returns:
        A `AuthorizedSession` object or None: Retrived Google Cloud
        AuthorizedSession or None if failed to retrive credentials.
    """
    logger = logging.getLogger("primary_logger")
    credentials = get_google_credentials(secret_name)
    if credentials is None:
        return None
    try:
        session = AuthorizedSession(credentials)
        return session
    except Exception as e:
        logger.exception(f"Get Google Cloud AuthorizedSession error with {str(e)}")
        return None


def get_general_credentials(secret_name: str) -> Union[str, None]:
    """
    Retrieves a secret value from the environment variables.

    This function retrieves a secret value from the environment variables
    by calling the `os.getenv()` function with the appropriate environment
    variable name. The function then returns the resulting secret value as
    a string.

    Args:
        secret_name (str): The name of the environment variable containing
        the secret value.

    Returns:
        str or None: The secret value as a string or None if failed to
        retrieve secret value.
    """
    logger = logging.getLogger("primary_logger")
    secret_value_encoded = os.getenv(secret_name)
    if secret_value_encoded is None:
        logger.error(f"Failed to get secret value with {secret_name}")
        return None
    secret_value = base64.b64decode(secret_value_encoded).decode("utf-8")
    return secret_value



import logging
import requests
from requests.models import Response
from unittest.mock import Mock, patch
from cru_dse_utils import get_request


# Test get_request for a successful request where everything works as expected.
def test_get_request():
    mock_response = Mock(spec=Response)
    mock_response.status_code = 200
    mock_response.json.return_value = {"key": "value"}

    logger = logging.getLogger("test")
    url = "https://test.com"
    headers = {"Header": "test"}
    params = {"Param": "test"}

    with patch("requests.get", return_value=mock_response) as mock_get:
        result = get_request(url, headers, params)

    assert result == mock_response
    mock_get.assert_called_once_with(url, headers=headers, params=params, timeout=60)


# Test get_request for timeout error.
def test_get_request_timeout_error():
    logger = logging.getLogger("test")
    url = "https://test.com"
    headers = {"Header": "test"}
    params = {"Param": "test"}

    with patch("requests.get", side_effect=requests.exceptions.Timeout()), patch(
        "time.sleep"
    ):
        assert get_request(url, headers, params) is None


# Test get_request for connection error.
def test_get_request_json_error():
    mock_response = Mock(spec=Response)
    mock_response.status_code = 200
    mock_response.json.side_effect = ValueError()

    logger = logging.getLogger("test")
    url = "https://test.com"
    headers = {"Header": "test"}
    params = {"Param": "test"}

    with patch("requests.get", return_value=mock_response), patch(
        "time.sleep", return_value=None
    ):
        assert get_request(url, headers, params) is None
